library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity lin_uart is
  port (
    rst, clk : in    std_logic;
    -- lin inputs and outputs
    lin_data     : inout std_logic_vector(7 downto 0);
    lin_addr     : out   std_logic_vector(3 downto 0);
    lin_seln     : out   std_logic;
    lin_rn       : out   std_logic;
    lin_wn       : out   std_logic;
    lin_int      : in    std_logic;
    -- Pass-through to provide voltage translation
    lin_TxD      : in    std_logic;
    lin_RxD      : out   std_logic;
    -- To LIN transceiver
    lin_xcvr_TxD     : out   std_logic;
    lin_xcvr_RxD     : in    std_logic;
    -- To RS232 transceiver
    rs232_TxD : out std_logic;
    rs232_RxD : in  std_logic;
    rs232_RTS : out std_logic;
    rs232_CTS : in std_logic
    );
end lin_uart;

architecture Mixed of lin_uart is

  component tx
    port (
      rst       : in  std_logic;
      clk_x16   : in  std_logic;
      data_in   : in  std_logic_vector (7 downto 0);
      TxD       : out std_logic;
      data_sent : out std_logic;
      strobe    : in  std_logic
      );
  end component;

  component rcv
    port (
      rst        : in  std_logic;
      clk_x16    : in  std_logic;
      data_out   : out std_logic_vector (7 downto 0);
      RxD        : in  std_logic;
      data_ready : out std_logic;
      strobe     : in  std_logic
      );
  end component;

  type uart_state_t is (
    s_addr,                            -- Nothing pending
    s_wait_reading,
    s_reading,
    s_latch_xmit,
    s_wait_xmit,
    s_wait_writing,
    s_wait_addr,
    s_writing
  );

  signal uart_state : uart_state_t;
  signal addr : std_logic_vector (3 downto 0);
  signal data_xmit, data_rcv : std_logic_vector (7 downto 0);
  signal xmit_data_sent, rcv_data_ready : std_logic;
  signal tx_strobe, rcv_strobe : std_logic;
  signal lin_rn_internal, lin_wn_internal, data_valid : std_logic;
  signal clk_x16 : std_logic;
  signal rcv_data_ready_last : std_logic;
  signal clkcounter : std_logic_vector (2 downto 0);

begin
  tx1 : tx port map (
      rst       => rst,
      clk_x16   => clk_x16,
      data_in   => data_xmit,
      TxD       => rs232_TxD,
      data_sent => xmit_data_sent,
      strobe    => tx_strobe
      );

  rcv1 : rcv port map (
      rst        => rst,
      clk_x16    => clk_x16,
      data_out   => data_rcv,
      RxD        => rs232_RxD,
      data_ready => rcv_data_ready,
      strobe     => rcv_strobe
      );

--  divide: process (rst, clk)
--  begin
--    if (rst = '1') then
--      clkcounter <= "001";
--      clk_x16 <= '0';
--    elsif (rising_edge(clk)) then
--      clkcounter <= clkcounter + 1;
--      -- 19200
----      if (clkcounter = 6) then
----      -- 57600
----      if (clkcounter = 2) then
--        clkcounter <= "001";
--        clk_x16 <= '1';
--      else
--        clk_x16 <= '0';
--      end if;
--    end if;
--  end process divide;
  -- 115200
  clk_x16 <= clk;

  process (clk_x16, rst)
  begin
    if (rst = '1') then
      addr <= "0000";
      uart_state <= s_addr;
      rcv_strobe <= '0';
      tx_strobe <= '0';
      rcv_data_ready_last <= '0';
    elsif (rising_edge(clk_x16)) then
      rcv_data_ready_last <= rcv_data_ready;

      case uart_state is
        when s_addr  =>
          rcv_strobe <= '0';
          tx_strobe <= '0';
          -- Expecting an address byte to be input
          if (rcv_data_ready = '1') then
            -- Reset the uart receiver
            rcv_strobe <= '1';
            -- Address byte uses a special protocol:
            -- bit 7 is the state of the read/write flags.
            -- Bits 3-0 are the address of the register to write/read to/from.
            addr <= data_rcv(3 downto 0);
            if (data_rcv(7) = '0') then
              uart_state <= s_wait_reading;
            else
              uart_state <= s_wait_writing;
            end if;
          else
            uart_state <= s_addr;
          end if;

        when s_wait_reading =>
          -- Implied: rn = 0 (reading from lin controller)
            uart_state <= s_reading;

        when s_reading =>
            uart_state <= s_latch_xmit;
            tx_strobe <= '1';

        when s_latch_xmit =>
          if (xmit_data_sent = '0') then
            tx_strobe <= '0';
            uart_state <= s_wait_xmit;
          end if;
          
        when s_wait_xmit =>
          -- Wait until the data's sent through the UART
          if (xmit_data_sent = '0') then
            uart_state <= s_wait_xmit;
          else 
            uart_state <= s_addr;
          end if;

        when s_wait_writing =>
          rcv_strobe <= '0';
          uart_state <= s_writing;
          
        when s_writing =>
          rcv_strobe <= '0';
          -- Expecting a data byte to be input
          if (rcv_data_ready = '1') then
            -- Reset the uart receiver
            rcv_strobe <= '1';
            -- writing to the lin controller, reading from the uart
            uart_state <= s_wait_addr;
          else
            uart_state <= s_writing;
          end if;
          
        when s_wait_addr =>
          rcv_strobe <= '0';
          uart_state <= s_addr;
          
        when others =>
          uart_state <= s_addr;
      end case;

    end if;
  end process;

  process (clk_x16, rst)
  begin
    if (rst = '1') then
      data_valid <= '0';
    elsif (rising_edge(clk_x16)) then
      data_valid <= not(lin_wn_internal);
    end if;
  end process;

  lin_rn_internal <= '0' when (uart_state = s_reading or
                               uart_state = s_latch_xmit) else
                     '1';
  lin_wn_internal <= '0' when (uart_state = s_writing) else '1';
  lin_wn <= lin_wn_internal;
  lin_rn <= lin_rn_internal;
  lin_seln <= (lin_rn_internal and lin_wn_internal);

  lin_addr <= addr;
  lin_xcvr_TxD <= lin_TxD;
  lin_RxD <= lin_xcvr_RxD;
  
  lin_data <= data_rcv when (data_valid ='1') else "ZZZZZZZZ";
  data_xmit <= lin_data when (tx_strobe = '1') else "00000000";

  -- Just loop back RTS/CTS
  rs232_RTS <= rs232_CTS;
  
  
end Mixed;
